@@ -23,7 +23,7 @@
 http://jackclient-python.readthedocs.io/
 
 """
-__version__ = '0.4.6'
+__version__ = '0.5.0'
 
 from ctypes.util import find_library as _find_library
 import errno as _errno
@@ -56,10 +56,32 @@ STARTING = _lib.JackTransportStarting
 NETSTARTING = _lib.JackTransportNetStarting
 """Waiting for sync ready on the network."""
 
+PROPERTY_CREATED = _lib.PropertyCreated
+"""A property was created.  See `Client.set_property_change_callback()`."""
+PROPERTY_CHANGED = _lib.PropertyChanged
+"""A property was changed.  See `Client.set_property_change_callback()`."""
+PROPERTY_DELETED = _lib.PropertyDeleted
+"""A property was deleted.  See `Client.set_property_change_callback()`."""
+
 _SUCCESS = 0
 _FAILURE = 1
 
 
+def _decode(cdata):
+    return _ffi.string(cdata).decode()
+
+
+# Get metadata constants from library
+for name in dir(_lib):
+    if name.startswith('JACK_METADATA_'):
+        try:
+            globals()[name[5:]] = _decode(getattr(_lib, name))
+        except _ffi.error:
+            pass
+else:
+    del name
+
+
 class Client(object):
     """A client that can connect to the JACK audio server."""
 
@@ -146,7 +168,15 @@ class Client(object):
     @property
     def name(self):
         """The name of the JACK client (read-only)."""
-        return _ffi.string(_lib.jack_get_client_name(self._ptr)).decode()
+        return _decode(_lib.jack_get_client_name(self._ptr))
+
+    @property
+    def uuid(self):
+        """The UUID of the JACK client (read-only)."""
+        uuid = _ffi.gc(_lib.jack_client_get_uuid(self._ptr), _lib.jack_free)
+        if not uuid:
+            raise JackError('Unable to get UUID')
+        return _decode(uuid)
 
     @property
     def samplerate(self):
@@ -510,11 +540,9 @@ class Client(object):
         effect in two process cycles.  If there are slow-sync clients
         and the transport is already rolling, it will enter the
         `STARTING` state and begin invoking their sync callbacks
-        (see `jack_set_sync_callback()`__) until ready.
+        (see `set_sync_callback()`) until ready.
         This function is realtime-safe.
 
-        __ http://jackaudio.org/files/docs/html/group__TransportControl.html
-
         Parameters
         ----------
         position : jack_position_t
@@ -529,6 +557,29 @@ class Client(object):
         _check(_lib.jack_transport_reposition(self._ptr, position),
                'Error re-positioning transport')
 
+    def set_sync_timeout(self, timeout):
+        """Set the timeout value for slow-sync clients.
+
+        This timeout prevents unresponsive slow-sync clients from
+        completely halting the transport mechanism.  The default is two
+        seconds.  When the timeout expires, the transport starts
+        rolling, even if some slow-sync clients are still unready.
+        The *sync callbacks* of these clients continue being invoked,
+        giving them a chance to catch up.
+
+        Parameters
+        ----------
+        timeout : int
+            Delay (in microseconds) before the timeout expires.
+
+        See Also
+        --------
+        set_sync_callback
+
+        """
+        _check(_lib.jack_set_sync_timeout(self._ptr, timeout),
+               'Error setting sync timeout')
+
     def set_freewheel(self, onoff):
         """Start/Stop JACK's "freewheel" mode.
 
@@ -605,7 +656,7 @@ class Client(object):
         """
         @self._callback('JackInfoShutdownCallback')
         def callback_wrapper(code, reason, _):
-            callback(Status(code), _ffi.string(reason).decode())
+            callback(Status(code), _decode(reason))
 
         _lib.jack_on_info_shutdown(self._ptr, callback_wrapper, _ffi.NULL)
 
@@ -844,7 +895,7 @@ class Client(object):
         """
         @self._callback('JackClientRegistrationCallback')
         def callback_wrapper(name, register, _):
-            callback(_ffi.string(name).decode(), bool(register))
+            callback(_decode(name), bool(register))
 
         _check(_lib.jack_set_client_registration_callback(
             self._ptr, callback_wrapper, _ffi.NULL),
@@ -1055,8 +1106,7 @@ class Client(object):
             else:
                 port = None
             try:
-                callback(port, _ffi.string(old_name).decode(),
-                         _ffi.string(new_name).decode())
+                callback(port, _decode(old_name), _decode(new_name))
             except CallbackExit:
                 return _FAILURE
             return _SUCCESS
@@ -1146,6 +1196,67 @@ class Client(object):
             self._ptr, callback_wrapper, _ffi.NULL),
             'Error setting xrun callback')
 
+    def set_sync_callback(self, callback):
+        """Register (or unregister) as a slow-sync client.
+
+        A slow-sync client is one that cannot respond immediately to
+        transport position changes.
+
+        The *callback* will be invoked at the first available
+        opportunity after its registration is complete.  If the client
+        is currently active this will be the following process cycle,
+        otherwise it will be the first cycle after calling `activate()`.
+        After that, it runs whenever some client requests a new
+        position, or the transport enters the `STARTING` state.
+        While the client is active, this callback is invoked just before
+        the *process callback* (see `set_process_callback()`) in the
+        same thread.
+
+        Clients that don't set a *sync callback* are assumed to be ready
+        immediately any time the transport wants to start.
+
+        Parameters
+        ----------
+        callback : callable or None
+
+            User-supplied function that returns ``True`` when the
+            slow-sync client is ready.  This realtime function must not
+            wait.  It must have this signature::
+
+                callback(state: int, pos: jack_position_t) -> bool
+
+            The *state* argument will be:
+
+            - `STOPPED` when a new position is requested;
+            - `STARTING` when the transport is waiting to start;
+            - `ROLLING` when the timeout has expired, and the position
+              is now a moving target.
+
+            The *pos* argument holds the new transport position using
+            the same structure as returned by
+            `transport_query_struct()`.
+
+            Setting *callback* to ``None`` declares that this
+            client no longer requires slow-sync processing.
+
+        See Also
+        --------
+        set_sync_timeout
+
+        """
+        if callback is None:
+            callback_wrapper = _ffi.NULL
+        else:
+
+            @self._callback('JackSyncCallback', error=False)
+            def callback_wrapper(state, pos, _):
+                return callback(state, pos)
+
+        _check(
+            _lib.jack_set_sync_callback(
+                self._ptr, callback_wrapper, _ffi.NULL),
+            'Error setting sync callback')
+
     def set_timebase_callback(self, callback=None, conditional=False):
         """Register as timebase master for the JACK subsystem.
 
@@ -1226,6 +1337,35 @@ class Client(object):
         _check(err, 'Error setting timebase callback')
         return True
 
+    def set_property_change_callback(self, callback):
+        """Register property change callback.
+
+        Tell the JACK server to call *callback* whenever a property is
+        created, changed or deleted.
+
+        Parameters
+        ----------
+        callback : callable
+            User-supplied function that is called whenever a property is
+            created, changed or deleted.  It must have this signature::
+
+                callback(subject: int, key: str, change: int) -> None
+
+            The first and second arguments are the *subject* and *key*,
+            respectively.  See `set_property()` for details.
+            The third argument has one of the values `PROPERTY_CREATED`,
+            `PROPERTY_CHANGED` or `PROPERTY_DELETED`, which should be
+            self-explanatory.
+
+        """
+        @self._callback('JackPropertyChangeCallback')
+        def callback_wrapper(subject, key, change, _):
+            callback(subject, _decode(key) if key else '', change)
+
+        _check(_lib.jack_set_property_change_callback(
+            self._ptr, callback_wrapper, _ffi.NULL),
+            'Error setting property change callback')
+
     def get_uuid_for_client_name(self, name):
         """Get the session ID for a client name.
 
@@ -1237,7 +1377,7 @@ class Client(object):
             self._ptr, name.encode()), _lib.jack_free)
         if not uuid:
             raise JackError('Unable to get session ID for {0!r}'.format(name))
-        return _ffi.string(uuid).decode()
+        return _decode(uuid)
 
     def get_client_name_by_uuid(self, uuid):
         """Get the client name for a session ID.
@@ -1250,7 +1390,7 @@ class Client(object):
             self._ptr, uuid.encode()), _lib.jack_free)
         if not name:
             raise JackError('Unable to get client name for {0!r}'.format(uuid))
-        return _ffi.string(name).decode()
+        return _decode(name)
 
     def get_port_by_name(self, name):
         """Get port by name.
@@ -1328,6 +1468,141 @@ class Client(object):
             _lib.jack_free)
         return self._port_list_from_pointers(names)
 
+    def set_property(self, subject, key, value, type=''):
+        """Set a metadata property on *subject*.
+
+        Parameters
+        ----------
+        subject : int or str
+            The subject (UUID) to set the property on.
+            UUIDs can be obtained with `Client.uuid`, `Port.uuid` and
+            `Client.get_uuid_for_client_name()`.
+        key : str
+            The key (URI) of the property.  Some predefined keys are
+            available as ``jack.METADATA_*`` module constants.
+        value : str or bytes
+            The value of the property.
+        type : str, optional
+            The type of the property, either a MIME type or URI.
+            If *type* is empty, the *value* is assumed to be a UTF-8
+            encoded string (``'text/plain'``).
+
+            Example values:
+
+            - ``'image/png;base64'`` (base64 encoded PNG image)
+            - ``'http://www.w3.org/2001/XMLSchema#int'`` (integer)
+
+            Official types are preferred, but clients may use any
+            syntactically valid MIME type (which start with a type and
+            slash, like ``'text/...'``).  If a URI type is used, it must
+            be a complete absolute URI (which start with a scheme and
+            colon, like ``'http:'``).
+
+        See Also
+        --------
+        get_property
+        get_properties
+        get_all_properties
+        remove_property
+        remove_properties
+        remove_all_properties
+        set_property_change_callback
+
+        """
+        subject = _uuid_parse(subject)
+        if isinstance(value, str):
+            value = value.encode()
+        if isinstance(type, str):
+            type = type.encode()
+        if _lib.jack_set_property(
+                self._ptr, subject, key.encode(), value, type) != 0:
+            raise ValueError('Unable to set property {!r} for subject {!r}'
+                             .format(key, subject))
+
+    def remove_property(self, subject, key):
+        """Remove a single metadata property on *subject*.
+
+        Parameters
+        ----------
+        subject : int or str
+            The subject (UUID) to remove the property from.
+            UUIDs can be obtained with `Client.uuid`, `Port.uuid` and
+            `Client.get_uuid_for_client_name()`.
+        key : str
+            The key of the property to be removed.
+
+        See Also
+        --------
+        set_property
+        get_property
+        get_properties
+        get_all_properties
+        remove_properties
+        remove_all_properties
+        set_property_change_callback
+
+        """
+        subject = _uuid_parse(subject)
+        if _lib.jack_remove_property(self._ptr, subject, key.encode()) != 0:
+            raise ValueError('Unable to remove property {!r} for subject {!r}'
+                             .format(key, subject))
+
+    def remove_properties(self, subject):
+        """Remove all metadata properties on *subject*.
+
+        Parameters
+        ----------
+        subject : int or str
+            The subject (UUID) to remove all properties from.
+            UUIDs can be obtained with `Client.uuid`, `Port.uuid` and
+            `Client.get_uuid_for_client_name()`.
+
+        Returns
+        -------
+        int
+            The number of properties removed.
+
+        See Also
+        --------
+        set_property
+        get_property
+        get_properties
+        get_all_properties
+        remove_property
+        remove_all_properties
+        set_property_change_callback
+
+        """
+        subject = _uuid_parse(subject)
+        number = _lib.jack_remove_properties(self._ptr, subject)
+        if number < 0:
+            raise RuntimeError(
+                'Unable to remove properties for subject {!r}'.format(subject))
+        return number
+
+    def remove_all_properties(self):
+        """Remove all metadata properties.
+
+        .. warning::
+
+            This deletes all metadata managed by a running JACK server.
+            Data lost cannot be recovered (though it can be recreated by
+            new calls to `set_property()`).
+
+        See Also
+        --------
+        set_property
+        get_property
+        get_properties
+        get_all_properties
+        remove_property
+        remove_properties
+        set_property_change_callback
+
+        """
+        if _lib.jack_remove_all_properties(self._ptr) != 0:
+            raise RuntimeError('Unable to remove properties')
+
     def _callback(self, cdecl, **kwargs):
         """Wrapper for ffi.callback() that keeps callback alive."""
         def callback_decorator(python_callable):
@@ -1358,7 +1633,7 @@ class Client(object):
                 name = names[idx]
                 if not name:
                     break
-                ports.append(self.get_port_by_name(_ffi.string(name).decode()))
+                ports.append(self.get_port_by_name(_decode(name)))
                 idx += 1
         return ports
 
@@ -1425,7 +1700,7 @@ class Port(object):
     @property
     def name(self):
         """Full name of the JACK port (read-only)."""
-        return _ffi.string(_lib.jack_port_name(self._ptr)).decode()
+        return _decode(_lib.jack_port_name(self._ptr))
 
     @property
     def shortname(self):
@@ -1438,7 +1713,7 @@ class Port(object):
         `port_name_size()`, it will be truncated.
 
         """
-        return _ffi.string(_lib.jack_port_short_name(self._ptr)).decode()
+        return _decode(_lib.jack_port_short_name(self._ptr))
 
     @shortname.setter
     def shortname(self, shortname):
@@ -1454,7 +1729,7 @@ class Port(object):
         result = []
         if _lib.jack_port_get_aliases(self._ptr, aliasesptr) > 0:
             for i in 0, 1:
-                alias = _ffi.string(aliases[i]).decode()
+                alias = _decode(aliases[i])
                 if alias:
                     result.append(alias)
 
@@ -2356,6 +2631,143 @@ class CallbackExit(Exception):
     pass
 
 
+def _uuid_parse(uuid):
+    if isinstance(uuid, (Client, Port)):
+        uuid = uuid.uuid
+
+    if isinstance(uuid, int):
+        return uuid
+    elif isinstance(uuid, str):
+        uuid_ptr = _ffi.new('jack_uuid_t*')
+        if _lib.jack_uuid_parse(uuid.encode(), uuid_ptr) != 0:
+            raise ValueError('Unable to parse UUID: {!r}'.format(uuid))
+        return uuid_ptr[0]
+    raise TypeError('Invalid UUID: {!r}'.format(uuid))
+
+
+def _description_to_dict(desc):
+    assert desc != _ffi.NULL
+    prop_dict = {}
+    for i in range(desc.property_cnt):
+        prop = desc.properties[i]
+        key = _decode(prop.key)
+        prop_dict[key] = (
+            _ffi.string(prop.data),
+            _decode(prop.type) if prop.type else '')
+    free_description_itself = 0
+    _lib.jack_free_description(desc, free_description_itself)
+    return prop_dict
+
+
+def get_property(subject, key):
+    """Get a metadata property on *subject*.
+
+    Parameters
+    ----------
+    subject : int or str
+        The subject (UUID) to get the property from.
+        UUIDs can be obtained with `Client.uuid`, `Port.uuid` and
+        `Client.get_uuid_for_client_name()`.
+    key : str
+        The key of the property.
+
+    Returns
+    -------
+    (bytes, str) or None
+        A tuple containing the value of the property and the type of the
+        property.  If *subject* doesn't have the property *key*,
+        ``None`` is returned.
+
+    See Also
+    --------
+    Client.set_property
+    get_properties
+    get_all_properties
+    Client.remove_property
+    Client.remove_properties
+    Client.remove_all_properties
+    set_property_change_callback
+
+    """
+    subject = _uuid_parse(subject)
+    data = _ffi.new('char**')
+    type = _ffi.new('char**')
+    if _lib.jack_get_property(subject, key.encode(), data, type) != 0:
+        return None
+    data = _ffi.gc(data[0], _lib.jack_free)
+    type = _ffi.gc(type[0], _lib.jack_free)
+    return _ffi.string(data), _decode(type) if type else ''
+
+
+def get_properties(subject):
+    """Get all metadata properties of *subject*.
+
+    Parameters
+    ----------
+    subject : int or str
+        The subject (UUID) to get all properties of.
+        UUIDs can be obtained with `Client.uuid`, `Port.uuid` and
+        `Client.get_uuid_for_client_name()`.
+
+    Returns
+    -------
+    dict
+        A dictionary mapping property names to ``(value, type)`` tuples.
+
+    See Also
+    --------
+    Client.set_property
+    get_property
+    get_all_properties
+    Client.remove_property
+    Client.remove_properties
+    Client.remove_all_properties
+    set_property_change_callback
+
+    """
+    subject = _uuid_parse(subject)
+    desc = _ffi.new('jack_description_t*')
+    number = _lib.jack_get_properties(subject, desc)
+    if number < 0:
+        raise RuntimeError(
+            'Unable to get properties for subject {!r}'.format(subject))
+    assert number == desc.property_cnt
+    return _description_to_dict(desc)
+
+
+def get_all_properties():
+    """Get all properties for all subjects with metadata.
+
+    Returns
+    -------
+    dict
+        A dictionary mapping UUIDs to nested dictionaries as returned by
+        `get_properties()`.
+
+    See Also
+    --------
+    Client.set_property
+    get_property
+    get_properties
+    Client.remove_property
+    Client.remove_properties
+    Client.remove_all_properties
+    set_property_change_callback
+
+    """
+    descs = _ffi.new('jack_description_t**')
+    number = _lib.jack_get_all_properties(descs)
+    if number < 0:
+        raise RuntimeError('Error getting all properties')
+    descs = _ffi.gc(descs[0], _lib.jack_free)
+    prop_dict = {}
+    for idx in range(number):
+        desc = descs + idx
+        assert desc.subject not in prop_dict
+        prop_dict[desc.subject] = _description_to_dict(desc)
+    return prop_dict
+
+
 def position2dict(pos):
     """Convert CFFI position struct to a dict."""
     assert pos.unique_1 == pos.unique_2
@@ -2385,7 +2797,7 @@ def version():
 
 def version_string():
     """Get human-readable JACK version."""
-    return _ffi.string(_lib.jack_get_version_string()).decode()
+    return _decode(_lib.jack_get_version_string())
 
 
 def client_name_size():
@@ -2460,7 +2872,7 @@ def _set_error_or_info_function(callback, setter):
     else:
         @_ffi.callback('void (*)(const char*)')
         def callback_wrapper(msg):
-            callback(_ffi.string(msg).decode())
+            callback(_decode(msg))
 
         _keepalive[setter] = callback_wrapper
     setter(callback_wrapper)
